( lib-fake -- fake object handling library -- by Jaffa Tamarin )
 
( This library allows you to include support for fake objects in your
( programs with minimal alterations to your code.  In many cases all you
( need do is $include $lib/fake at the beginning of your program, and
( replace the MUF primitives you use for object handling with the versions
( below (replace match with fmatch, location with flocation, etc)
 
( If you $include $lib/quickfake instead of $lib/fake, the new versions
( of the primitives will be used automatically (see @view $lib/quickfake)
 
( The following MUF primitives are supported:
(   fmatch, flocation, fcontrols, fsetname, frecycle, fmoveto, funparseobj
(   fowner, fname, fdesc, fsetdesc, fcontents, fnext, fok?, fthing?, fdbcmp
(   faddprop, fsetprop, fgetprop, fgetpropstr, fgetpropval, fremove_prop
 
( These new functions are defined:
(   fake? (d -- b)
(       - returns true if object is fake
(   newfake (d s -- d)
(       - takes location and name and creates a new fake object
(   fdark? (d -- b)
(       - returns true if item is dark
(   fsetdark (d b -- or d s -- )
(       - sets an object dark if b is true, or !dark if b is false
(       - if the second parameter is a string and the object is fake,
(         the string will be copied into the object's 'show' field
(   fshow (d -- s)
(       - returns string value of fake object's 'show' field
(   fcheck-mpi (d -- b)
(       - returns true if fake object contains any MPI code
(         (always false for a real object)
 
(   noisy_fmatch (s -- d)
(       - as fmatch, but gives an error message if match fails
(   match_trigger
(       - if match is successful, returns the object in 'trigger @'
(       - and 1 on stack, otherwise returns <error-string> 0 on stack
(   match_contents (d s -- d)
(       - matches string against contents of location
(   fake_object (d s -- d)
(       - checks for fake object with given name in given location
(       - returns object found, or #-1 if no matching fake object
  
( Notes:
(       - moveto on a fake object requires on M3
(       - functions which create or alter a fake object require an M3,
(         or for the user to control the object location
(       - only first level props are supported on fake objects
(       - these routines use the first two 'var' slots )
 
$def version prog "_lib-version" getpropstr
 
$def old-lib-fake ( define to include support for fgetprop and fsetprop )
                  ( from 1.x versions of $lib/fake                      )
 
$ifdef j-cmd-look
$include $lib/look-install
$def fake-show-default $ifdef fake-show-default-is-no "no" $else "yes" $endif
$undef __glowver
$endif
 
( Edit these defines to match your fake object implementation      )
( Defaults are suitable for Jaffa's cmd-look or Mystique's NewLook )
( On GlowMuck, default installation is compatible with @objects    )
 
( prop containing description for fake object %n )
$ifdef __glowver
$def fake-desc "/_obj/%n"
$else
$def fake-desc "/_fake/%n/desc"
$endif
( set this define to 0 if you don't support alias names for fake objects   )
( e.g. a fake object called "stuffed toy fox;plushie" would appear as just )
( "stuffed toy fox" but would match "plushie" as an alias for the name     )
$define fake-aliases 1 $enddef
 
( set this define to 0 if fake objects are implemented with just a )
( description -- e.g. @objects on GlowMuck -- instead of a propdir )
$ifdef __glowver
$define fake-use-propdir 0 $enddef
$else
$define fake-use-propdir 1 $enddef
$endif
 
$ifdef fake-use-propdir>0
( prop defining whether fake object is visible -- i.e. shows in contents )
$def fake-show "/_fake/%n/show"
 
( default value for above prop )
$ifndef j-cmd-look
$def fake-show-default "no"
$endif
( propdir containing other details for fake object %n )
$def fake-dir "/_fake/%n/"
 
$else
$undef fake-use-propdir
$endif
 
$ifdef fake-aliases>0
$include $lib/multiname
$else
$def multi-propdir? swap over propdir? if 1 else pop 0 then
$def multi-getpropstr swap over getpropstr dup not if swap pop then
$undef fake-aliases
$endif
 
$def r0 $ifdef jaffa-install-loc jaffa-install-loc $else #0 $endif
$def FB-perms $ifdef __glowver "compatible_muf" sysparm dup if "y*" smatch then $else 1 $endif
$def write-r0? r0 #0 dbcmp $ifdef __glowver prog mlevel 3 > FB-perms or and $endif
 
var f-dbref
var f-name
 
$def inside? over location over dbcmp rot rot dbcmp or
$def getuid caller "S" flag? if caller owner else "me" match then
$def streq stringcmp not
 
$def ftos f-dbref @ intostr " " strcat f-name @ strcat
$def stof dup if dup " " instr 1 - strcut 1 strcut f-name ! pop atoi dbref else pop #-1 then f-dbref !
 
$def new-id prog "next" getpropval dup if 1 - else pop -5 then dup prog "next" rot setprop dbref
$def update-id intostr ftos prog "id/" 4 pick strcat 3 pick setprop prog "db/" rot strcat rot setprop
$def find-id ftos prog "db/" 3 pick strcat getpropstr dup if swap prog "id/" 4 pick strcat getpropstr streq if atoi dbref else pop new-id dup update-id then else pop pop new-id dup update-id then
$def get-id find-id
$def remove-id pop
$def unpack-id prog "id/" rot intostr strcat getpropstr stof
 
: check-fake ( d s -- b )
$ifdef fake-use-propdir
   over fake-dir rot "%n/" subst
 ( dup strlen 1 - strcut pop ( remove trailing / )
   multi-propdir?
$else
   over fake-desc rot "%n" subst
   multi-getpropstr
$endif
   if
    dup "/" rinstr strcut f-name ! pop f-dbref ! 1
   else
    pop 0
   then
;
 
: fake_object ( d s -- <obj> )
   check-fake if get-id else #-1 then
;
 
: fake-ok? ( d -- b )
   dup int -5 > if pop 0 exit then
   unpack-id
   f-dbref @ ok? if
$ifdef fake-use-propdir
    f-dbref @ fake-dir f-name @ "%n" subst propdir?
$else
    f-dbref @ fake-desc f-name @ "%n" subst getpropstr if 1 else 0 then
$endif
   else 0 then
;
 
( : unpack-check ( d -- d b )
(    dup fake-ok? not if "Invalid fake object reference." abort then )
( ;                                                                  )
$def unpack-check ( )
 
: fdbcmp ( <obj> <obj> -- b )
   f-dbref @ f-name @ rot unpack-id f-name @ streq swap f-dbref @ dbcmp and
;
 
lvar length
: fsetname
   unpack-check
   caller mlevel 3 < if
    getuid
    f-dbref @ controls not if "SETNAME: Permission denied for fake object." abort then
   then
$ifdef fake-use-propdir
   f-dbref @ fake-dir f-name @ "%n" subst
   dup strlen length !
   ( <obj> s db prop )
   begin
    nextprop dup
   while
    f-dbref @ over getprop f-dbref @ 3 pick remove_prop
    over length @ strcut swap pop
    f-dbref @ fake-dir 6 pick "%n" subst rot strcat rot setprop
    f-dbref @ swap
   repeat
   ( <obj> s "" )
   pop f-name !
$else
   f-dbref @ fake-desc f-name @ "%n" subst
   over over getpropstr rot rot remove_prop
   f-dbref @ fake-desc 4 pick "%n" subst rot setprop
   f-name !
$endif
   update-id
;
 
: frecycle ( <obj> -- )
   unpack-check
   caller mlevel 3 < if
    getuid
    f-dbref @ controls not if "RECYCLE: Permission denied for fake object." abort then
   then
$ifdef fake-use-propdir
   f-dbref @ fake-dir f-name @ "%n" subst
   dup strlen 1 - strcut pop remove_prop
$else
   f-dbref @ fake-desc f-name @ "%n" subst remove_prop
$endif
   remove-id
;
 
: fcheck-mpi ( <obj> -- b )
   unpack-check pop
$ifdef fake-use-propdir
   f-dbref @ fake-dir f-name @ "%n" subst
   begin
    nextprop dup
   while
    f-dbref @ over getpropstr "{" instr if pop 1 exit then
    f-dbref @ swap
   repeat
   pop 0
$else
   f-dbref @ fake-desc f-name @ "%n" subst
   getpropstr "{" instr
$endif
;
 
: fmoveto ( <obj> d -- )
   caller mlevel 3 < if "MOVETO: Mucker level 3 required to move fake object." abort then
   dup ok? not if "MOVETO: Invalid destination for fake object." abort then
   swap unpack-check
$ifdef fake-use-propdir
   f-dbref @ fake-dir f-name @ "%n" subst
   ( tgt <obj> db prop )
   begin
    nextprop dup
   while
    ( tgt <obj> prop )
    f-dbref @ over getprop f-dbref @ 3 pick remove_prop
    ( tgt <obj> prop val )
    4 pick 3 pick rot setprop
    f-dbref @ swap
   repeat
   pop swap f-dbref !
$else
   over fake-desc f-name @ "%n" subst
   f-dbref @ over getpropstr f-dbref @ 3 pick remove_prop setprop
   swap f-dbref !
$endif
   update-id
;
 
: fdark? ( <obj> -- b )
   unpack-check pop
$ifdef fake-use-propdir
   f-dbref @ fake-show f-name @ "%n" subst "$lib/fake" 1 parseprop
   dup not if pop fake-show-default then
   "n*" smatch
$else
   pop 0
$endif
;
 
: fcreate ( d s -- <obj> )
   over ok? not if "NEWFAKE: Invalid location." abort then
   caller mlevel 3 < if
    getuid
    3 pick controls not if "NEWFAKE: Permission denied." abort then
   then
$ifdef fake-use-propdir
   over fake-show 3 pick "%n" subst "yes" setprop
$else
   over fake-desc 3 pick "%n" subst "New fake object." setprop
$endif
   f-name ! f-dbref ! get-id
;
 
: nextfake ( d or <obj> -- <obj> or #-1 )
   caller mlevel 2 < if
    caller owner over controls not if pop #-1 exit then
   then
$ifdef fake-use-propdir
   dup ok? if
    dup fake-dir "" "%n/" subst
   else
    unpack-check pop
    f-dbref @ dup fake-dir f-name @ "%n/" subst
   then
 ( dup strlen 1 - strcut pop ( remove trailing / )
$else
   dup ok? if
    dup fake-desc "" "%n" subst
   else
    unpack-check pop
    f-dbref @ dup fake-desc f-name @ "%n" subst
   then
$endif
   nextprop
   dup not if pop pop #-1 exit then
   dup "/" rinstr strcut f-name ! pop f-dbref ! get-id
;
 
: public_contents_match ( s d -- <obj> )
   "me" match dup me !
   over controls not if
    dup me @ location inside? over me @ inside? or not if
     ( farlook only on objects controlled by player )
     pop pop #-1 exit
    then
   then
$ifdef j-cmd-look
   #-1 0 fake.muf "contents_match" call pop swap pop
   dup #-4 dbcmp if pop get-id then
$else
   over over swap rmatch
   dup #-3 dbcmp if pop pop pop #-1 exit then
   dup exit? if pop #-1 then
   dup ok? if
    dup location 3 pick dbcmp if swap pop swap pop exit then
   then
   pop swap fake_object
$endif
;
 
$ifndef j-cmd-look
: match_trigger
   dup dup if
    dup "{$*|#*|\\**}*" smatch if
     dup "#" 1 strncmp if match else .atodbref then
     dup ok? if
      dup loc @ inside? over me @ inside? or not if
       me @ over controls not if pop #-1 then
      then
     then
    else
     match
     dup #-3 dbcmp if pop me @ getlink then
    then
    dup ok? if trigger ! pop 1 exit then
    #-2 dbcmp if pop "I don't know which one you mean!" 0 exit then
    loc @ over fake_object dup #-1 dbcmp if pop me @ over fake_object then
    dup #-1 dbcmp not if trigger ! pop 1 exit then
   then
   pop pop "I don't see that here." 0
;
$endif
 
: public_match_trigger
   "me" match dup me ! location loc !
$ifdef j-cmd-look
   dup if
    dup "#-*" smatch if
     .atodbref
     dup fake-ok? if trigger ! 1 exit then
    else
     fake.muf "match_trigger" call
     dup if trigger @ #-4 dbcmp if get-id trigger ! then then
     exit
    then
   then
   pop "I don't see that here." 0
$else
   match_trigger
$endif
;
 
lvar save-trigger
: fmatch
   trigger @ save-trigger !
   public_match_trigger
   if trigger @ else pop #-1 then
   save-trigger @ trigger !
;
 
: noisy_fmatch
   trigger @ save-trigger !
   public_match_trigger
   if trigger @ else .tell #-1 then
   save-trigger @ trigger !
;
 
: mpi-install ( set up some stuff for macros.mpi )
$ifdef fake-use-propdir
   "{eval:{subst:{:1},%fake,{fsubs:" fake-dir "{:2},{:3}}}}" strcat strcat
$else
   "{null}"
$endif
   "{eval:\\{exec!:{fsubs:" fake-desc ",{:1}}\\}}" strcat strcat
   rot dup if
    dup "_msgmacs/fdesc" 4 rotate setprop
    "_msgmacs/f-action" rot setprop
   else pop then
;
 
: install
$ifdef fake-aliases
   "dup \";\" instr dup if 1 - strcut then pop"
$else
   "( )"
$endif
   prog "_defs/multi-1name" rot setprop
 
   "dup f-dbref \"" fake-desc "\" rot f-name \"%n\" subst" strcat strcat
   prog "_defs/f-desc" rot setprop
 
$ifdef fake-use-propdir
   "dup f-dbref \"" fake-show "\" rot f-name \"%n\" subst" strcat strcat
   prog "_defs/f-show" rot setprop
   "over f-dbref \"" fake-dir "\" 4 rotate f-name \"%n\" subst rot strcat" strcat strcat
   prog "_defs/makefakeprop" rot setprop
   prog "_defs/f-getprop" "makefakeprop \\getprop" setprop
   prog "_defs/f-getpropstr" "makefakeprop \\getpropstr" setprop
   prog "_defs/f-getpropval" "makefakeprop \\getpropval" setprop
   prog "_defs/f-setprop" "rot rot makefakeprop rot \\setprop" setprop
   prog "_defs/f-addprop" "4 rotate 4 rotate makefakeprop 4 rotate 4 rotate \addprop" setprop
   prog "_defs/f-remove_prop" "makefakeprop \\remove_prop" setprop
$else
   prog "_defs/show" "pop \"\"" setprop
   prog "_defs/fsetshow" "pop pop" setprop
   prog "_defs/f-getprop" "pop pop 0" setprop
   prog "_defs/f-getpropstr" "pop pop \"\"" setprop
   prog "_defs/f-getpropval" "pop pop 0" setprop
   prog "_defs/f-setprop" "pop pop pop" setprop
   prog "_defs/f-addprop" "pop pop pop pop" setprop
   prog "_defs/f-remove_prop" "pop pop" setprop
$endif
 
$ifdef old-lib-fake
   prog "_defs/fsetprop" getpropstr prog "_defs/new-fsetprop" rot setprop
   prog "_defs/fgetprop" getpropstr prog "_defs/new-fgetprop" rot setprop
   prog "_defs/fsetprop" "3 pick string? if old-fsetprop else new-fsetprop then" setprop
   prog "_defs/fgetprop" "over string? if old-fgetprop else new-fgetprop then" setprop
$endif
;
 
: main
   "Jaffa's fake object library, v" version strcat .tell
$ifdef j-cmd-look
   "Default installation for Jaffa's cmd-look" .tell
$endif
   install
   loc @ "_msgmacs/jaffa" envpropstr swap pop dup if
    dup " " rinstr strcut 2 strcut pop
    "v2" strcmp 0 < if
     "Warning: please install the latest version of macros.mpi" .tell
    then
   then pop
   prog "_defs/lib-fake" getpropstr if exit then
   prog "_defs/lib-fake" "#" prog intostr strcat setprop
   write-r0? if #0 else me @ then
   dup "_reg/lib/fake" prog setprop
   "_defs/j-lib-fake" version setprop
;
 
PUBLIC fdbcmp
PUBLIC fsetname
PUBLIC fmatch
PUBLIC fcreate
PUBLIC frecycle
PUBLIC fmoveto
PUBLIC fcheck-mpi
PUBLIC fake_object
PUBLIC noisy_fmatch
PUBLIC public_contents_match
PUBLIC public_match_trigger
PUBLIC nextfake
PUBLIC fdark?
PUBLIC mpi-install
PUBLIC fake-ok?

